import logging
from collections import defaultdict, namedtuple

from clint.textui import colored, indent, puts

from .utils import fullpath, bold, splitglob

logger = logging.getLogger(__name__)


Job = namedtuple("Job", "folderstr basedir path filters actions")
Job.__doc__ = """
    :param str folderstr: the original folder definition specified in the config
    :param Path basedir:  the job's base folder
    :param Path path:     the path of the file to handle
    :param list filters:  the filters that apply to the path
    :param list actions:  the actions which should be executed
"""


def all_files_for_rule(rule):
    files = {}
    for folderstr in rule.folders:
        exclude_flag = folderstr.startswith("!")
        basedir, globstr = splitglob(folderstr.lstrip("!"))
        if basedir.is_dir():
            if not globstr:
                globstr = "**/*" if rule.subfolders else "*"
        elif basedir.is_file():
            # this allows specifying single files
            globstr = basedir.name
            basedir = basedir.parent
        else:
            raise ValueError("Path not found: {}".format(folderstr))
        for path in basedir.glob(globstr):
            if path.is_file() and (
                rule.system_files
                or path.name not in ("thumbs.db", "desktop.ini", ".DS_Store")
            ):
                if not exclude_flag:
                    files[path] = (folderstr, basedir)
                elif path in files:
                    del files[path]
    for path, (folderstr, basedir) in files.items():
        yield (folderstr, basedir, path)


def find_jobs(rules):
    for rule in rules:
        for folderstr, basedir, path in all_files_for_rule(rule):
            if all(f.matches(path) for f in rule.filters):
                yield Job(
                    folderstr=folderstr,
                    basedir=basedir,
                    path=path,
                    filters=rule.filters,
                    actions=rule.actions,
                )


def group_by_folder(jobs):
    result = defaultdict(list)
    for job in jobs:
        result[job.folderstr].append(job)
    return result


def filter_pipeline(job):
    result = {}
    for filter_ in job.filters:
        result.update(filter_.parse(job.path))
    return result


def action_pipeline(job: Job, attrs: dict, simulate: bool):
    try:
        current_path = None
        for action in job.actions:
            if current_path is not None:
                attrs["path"] = current_path
            new_path = action.run(attrs=attrs, simulate=simulate)
            if new_path is not None:
                current_path = new_path
    except Exception as e:
        logger.exception(e)
        action.print("%s %s" % (colored.red("ERROR!", bold=True), e))


def execute_rules(rules, simulate):
    jobs = list(find_jobs(rules))
    if not jobs:
        msg = "Nothing to do."
        logger.info(msg)
        puts(msg)
        return

    jobs_by_folder = group_by_folder(jobs)
    first = True
    if simulate:
        puts(colored.green("SIMULATION ~~~", bold=True))
    for folder, jobs in sorted(jobs_by_folder.items()):
        # newline between folders
        if not first:
            puts()
        first = False

        puts("Folder %s:" % bold(folder))
        with indent(2):
            for job in jobs:
                path = job.path
                basedir = job.basedir
                relative_path = path.relative_to(basedir)
                puts("File %s:" % bold(relative_path))
                with indent(2):
                    attrs = filter_pipeline(job)
                    attrs["path"] = path
                    attrs["basedir"] = basedir
                    attrs["relative_path"] = relative_path
                    action_pipeline(job=job, attrs=attrs, simulate=simulate)
    if simulate:
        puts(colored.green("~~~ SIMULATION", bold=True))



import sys
from datetime import datetime, timedelta
from .filter import Filter


class Created(Filter):

    """
    Matches files by created date

    :param int days:
        specify number of days

    :param int hours:
        specify number of hours

    :param int minutes:
        specify number of minutes

    :param str mode:
        either 'older' or 'newer'. 'older' matches all files created before the given
        time, 'newer' matches all files created within the given time.
        (default = 'older')

    :returns:
        - ``{created.year}`` -- the year the file was created
        - ``{created.month}`` -- the month the file was created
        - ``{created.day}`` -- the day the file was created
        - ``{created.hour}`` -- the hour the file was created
        - ``{created.minute}`` -- the minute the file was created
        - ``{created.second}`` -- the second the file was created

    Examples:
        - Show all files on your desktop created at least 10 days ago:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - created:
                      days: 10
                actions:
                  - echo: 'Was created at least 10 days ago'

        - Show all files on your desktop which were created within the last 5 hours:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - created:
                      hours: 5
                      mode: newer
                actions:
                  - echo: 'Was created within the last 5 hours'

        - Sort pdfs by year of creation:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents'
                filters:
                  - extension: pdf
                  - created
                actions:
                  - move: '~/Documents/PDF/{created.year}/'
    """

    def __init__(self, days=0, hours=0, minutes=0, seconds=0, mode="older"):
        self._mode = mode.strip().lower()
        if self._mode not in ("older", "newer"):
            raise ValueError("Unknown option for 'mode': must be 'older' or 'newer'.")
        else:
            self.is_older = self._mode == "older"
        self.timedelta = timedelta(
            days=days, hours=hours, minutes=minutes, seconds=seconds
        )

    def matches(self, path):
        created_date = self._created(path)
        reference_date = datetime.now() - self.timedelta
        if self.is_older:
            return created_date <= reference_date
        else:
            return created_date >= reference_date

    def parse(self, path):
        created_date = self._created(path)
        return {"created": created_date}

    def _created(self, path):
        # see https://stackoverflow.com/a/39501288/300783
        stat = path.stat()
        if sys.platform.startswith('win'):
            time = stat.st_ctime
        else:
            try:
                time = stat.st_birthtime
            except AttributeError:
                # We're probably on Linux. No easy way to get creation dates here,
                # so we'll settle for when its content was last modified.
                time = stat.st_mtime
        return datetime.fromtimestamp(time)

    def __str__(self):
        return "Created(delta=%s, select_mode=%s)" % (self.timedelta, self._mode)



from typing import Optional
from clint.textui import puts
from organize.utils import Path


class Error(Exception):
    pass


class TemplateAttributeError(Error):
    pass


class Action:
    def run(self, attrs: dict, simulate: bool) -> Optional[Path]:
        # if you change the file path, return the new path here
        raise NotImplementedError

    def print(self, msg):
        """ print a message for the user """
        puts("- [%s] %s" % (self.__class__.__name__, msg))

    @staticmethod
    def fill_template_tags(msg: str, attrs: dict) -> str:
        try:
            return msg.format(**attrs)
        except AttributeError as exc:
            cause = exc.args[0]
            raise TemplateAttributeError(
                'Missing template variable %s for "%s"' % (cause, msg)
            )

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return "<%s>" % str(self)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



import inspect
import logging
import textwrap
from collections import namedtuple
from typing import List

import yaml

from . import actions, filters
from .utils import Path, first_key, flatten

logger = logging.getLogger(__name__)
Rule = namedtuple("Rule", "filters actions folders subfolders system_files")

# disable yaml constructors for strings starting with exclamation marks
# https://stackoverflow.com/a/13281292/300783
def default_yaml_constructor(loader, tag_suffix, node):
    return tag_suffix + " " + node.value


yaml.add_multi_constructor("", default_yaml_constructor)


class Config:
    def __init__(self, config: dict):  # type: (dict)
        self.config = config
        self.filter_by_name = {
            name.lower(): getattr(filters, name)
            for name, _ in inspect.getmembers(filters, inspect.isclass)
        }
        self.action_by_name = {
            name.lower(): getattr(actions, name)
            for name, _ in inspect.getmembers(actions, inspect.isclass)
        }

    @classmethod
    def parse_yaml(cls, config: str) -> dict:
        try:
            return yaml.safe_load(config)
        except yaml.YAMLError as e:
            raise cls.ParsingError(e)

    @classmethod
    def from_string(cls, config: str) -> "Config":
        dedented_config = textwrap.dedent(config)
        return cls(cls.parse_yaml(dedented_config))

    @classmethod
    def from_file(cls, path: Path) -> "Config":
        with path.open(encoding="utf-8") as f:
            return cls.from_string(f.read())

    def yaml(self) -> str:
        if not (self.config and "rules" in self.config):
            raise self.NoRulesFoundError()
        data = {"rules": self.config["rules"]}
        yaml.Dumper.ignore_aliases = lambda self, data: True
        return yaml.dump(
            data, allow_unicode=True, default_flow_style=False, default_style="'"
        )

    @staticmethod
    def parse_folders(rule_item):
        # the folder list is flattened so we can use encapsulated list
        # definitions in the config file.
        yield from flatten(rule_item["folders"])

    @staticmethod
    def sanitize_key(key):
        return key.lower().replace("_", "")

    def _get_filter_class_by_name(self, name):
        try:
            return self.filter_by_name[self.sanitize_key(name)]
        except AttributeError as e:
            raise self.Error("%s is no valid filter" % name) from e

    def _get_action_class_by_name(self, name):
        try:
            return self.action_by_name[self.sanitize_key(name)]
        except AttributeError as e:
            raise self.Error("%s is no valid action" % name) from e

    @staticmethod
    def _class_instance_with_args(Cls, args):
        if args is None:
            return Cls()
        elif isinstance(args, list):
            return Cls(*args)
        elif isinstance(args, dict):
            return Cls(**args)
        return Cls(args)

    def instantiate_filters(self, rule_item):
        # filter list can be empty
        try:
            filter_list = rule_item["filters"]
        except KeyError:
            return
        if not filter_list:
            return
        if not isinstance(filter_list, list):
            raise self.FiltersNoListError()

        for filter_item in filter_list:
            # filter with arguments
            if filter_item is None:
                yield None
            elif isinstance(filter_item, dict):
                name = first_key(filter_item)
                args = filter_item[name]
                filter_class = self._get_filter_class_by_name(name)
                yield self._class_instance_with_args(filter_class, args)
            # only given filter name without args
            elif isinstance(filter_item, str):
                name = filter_item
                filter_class = self._get_filter_class_by_name(name)
                yield filter_class()
            else:
                raise self.Error("Unknown filter: %s" % filter_item)

    def instantiate_actions(self, rule_item):
        action_list = rule_item["actions"]
        if not isinstance(action_list, list):
            raise self.ActionsNoListError()

        for action_item in action_list:
            if isinstance(action_item, dict):
                name = first_key(action_item)
                args = action_item[name]
                action_class = self._get_action_class_by_name(name)
                yield self._class_instance_with_args(action_class, args)
            elif isinstance(action_item, str):
                name = action_item
                action_class = self._get_action_class_by_name(name)
                yield action_class()
            else:
                raise self.Error("Unknown action: %s" % action_item)

    @property
    def rules(self) -> List[Rule]:
        """ :returns: A list of instantiated Rules """
        if not (self.config and "rules" in self.config):
            raise self.NoRulesFoundError()
        result = []
        for i, rule_item in enumerate(self.config["rules"]):
            # skip disabled rules
            if not rule_item.get("enabled", True):
                continue

            rule_folders = list(self.parse_folders(rule_item))
            rule_filters = list(self.instantiate_filters(rule_item))
            rule_actions = list(self.instantiate_actions(rule_item))

            if not rule_folders:
                logger.warning("No folders given for rule %s!", i + 1)
            if not rule_filters:
                logger.warning("No filters given for rule %s!", i + 1)
            if not rule_actions:
                logger.warning("No actions given for rule %s!", i + 1)

            rule = Rule(
                folders=rule_folders,
                filters=rule_filters,
                actions=rule_actions,
                subfolders=rule_item.get("subfolders", False),
                system_files=rule_item.get("system_files", False),
            )
            result.append(rule)
        return result

    class Error(Exception):
        pass

    class NoRulesFoundError(Error):
        def __str__(self):
            return "No rules found in configuration file"

    class ParsingError(Error):
        pass

    class NoFoldersFoundError(Error):
        pass

    class NoFiltersFoundError(Error):
        pass

    class NoActionsFoundError(Error):
        pass

    class FiltersNoListError(Error):
        def __str__(self):
            return "Please specify your filters as a YAML list"

    class ActionsNoListError(Error):
        def __str__(self):
            return "Please specify your actions as a YAML list"



import logging
from .action import Action


class Python(Action):

    r"""
    Execute python code in your config file.

    :param str code: The python code to execute

    Examples:
      - A basic example that shows how to get the current file path and do some
        printing in a for loop. The ``|`` is yaml syntax for defining a string
        literal spanning multiple lines.

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: '~/Desktop'
              actions:
                - python: |
                    print('The path of the current file is %s' % path)
                    for _ in range(5):
                        print('Heyho, its me from the loop')

      - You can access filter data:

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Desktop
              filters:
                - regex: '^(?P<name>.*)\.(?P<extension>.*)$'
              actions:
                - python: |
                    print('Name: %s' % regex.name)
                    print('Extension: %s' % regex.extension)

      - You have access to all the python magic -- do a google search for each
        filename starting with an underscore:

        .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders: ~/Desktop
              filters:
                - filename:
                    startswith: '_'
              actions:
                - python: |
                    import webbrowser
                    webbrowser.open('https://www.google.com/search?q=%s' % path.stem)
    """

    def __init__(self, code):
        self.code = code
        self.log = logging.getLogger(__name__)

    def run(self, attrs: dict, simulate: bool):
        if simulate:
            self.print("Code not run in simulation")
        else:
            path = attrs["path"]
            self.log.info(
                'Executing python script:\n"""\n%s""" with path="%s", args=%s',
                self.code,
                path,
                attrs,
            )
            # local variables for inline function
            locals_ = attrs.copy()
            locals_["simulate"] = simulate
            # replace default print function
            globals_ = globals().copy()
            globals_["print"] = self.print
            exec(self.code, globals_, locals_)



import logging

from .action import Action


class Trash(Action):

    """
    Move a file into the trash.

    Example:
        - Move all JPGs and PNGs on the desktop which are older than one year
          into the trash:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
              - filters:
                  - lastmodified:
                      - days: 365
                  - extension:
                      - png
                      - jpg
              - actions:
                  - trash
    """

    def __init__(self):
        self.log = logging.getLogger(__name__)

    def run(self, attrs: dict, simulate: bool):
        path = attrs["path"]
        from send2trash import send2trash

        self.print('Trash "%s"' % path)
        if not simulate:
            self.log.info("Moving file %s into trash.", path)
            send2trash(str(path))



import re
import sys
from collections import OrderedDict

import colorama

# in python < 3.6 the pathlib module misses some features so we have to import
# a backported alternative
if sys.version_info < (3, 6):
    from pathlib2 import Path
else:
    from pathlib import Path

unescaped_wildcard = re.compile(r"(?<!\\)[\*\?\[]+")


def splitglob(globstr):
    """ split a string with wildcards into a base folder and globstring """
    path = fullpath(globstr.strip())
    parts = path.parts
    for i, part in enumerate(parts):
        if unescaped_wildcard.search(part):
            return (Path(*parts[:i]), str(Path(*parts[i:])))
    return (path, "")


def fullpath(path):
    """ Expand '~' and resolve the given path """
    return Path(path).expanduser().resolve(strict=False)


def bold(text):
    # inspired by a feature request from the clint library
    # https://github.com/kennethreitz/clint/issues/157
    if sys.stdout.isatty():
        return "".join([colorama.Style.BRIGHT, str(text), colorama.Style.NORMAL])
    return text


def flatten(arr):
    if arr == []:
        return []
    if not isinstance(arr, list):
        return [arr]
    return flatten(arr[0]) + flatten(arr[1:])


def first_key(dic: dict):
    return list(dic.keys())[0]


class DotDict(OrderedDict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if not key.startswith("_OrderedDict__"):
            self[key] = value
        else:
            OrderedDict.__setattr__(self, key, value)

    def __delattr__(self, key):
        try:
            self.pop(key)
        except KeyError:
            OrderedDict.__delattr__(self, key)

    def __eq__(self, other):
        return dict.__eq__(self, other)

    def __str__(self):
        return "{%s}" % ", ".join("%r: %r" % (key, self[key]) for key in self)


def increment_filename_version(path: Path, separator=" ") -> Path:
    stem = path.stem
    try:
        # try to find any existing counter
        splitstem = stem.split(separator)  # raises ValueError on missing sep
        if len(splitstem) < 2:
            raise ValueError()
        counter = int(splitstem[-1])
        stem = separator.join(splitstem[:-1])
    except (ValueError, IndexError):
        # not found, we start with 1
        counter = 1
    return path.with_name(
        "{stem}{sep}{cnt}{suffix}".format(
            stem=stem, sep=separator, cnt=(counter + 1), suffix=path.suffix
        )
    )


def find_unused_filename(path: Path, separator=" ") -> Path:
    """
    We assume the given path already exists. This function adds a counter to the
    filename until we find a unused filename.
    """
    # TODO: Check whether the assumption can be eliminated for cleaner code.
    # TODO: Optimization: The counter only needs to be parsed once.
    tmp = path
    while True:
        tmp = increment_filename_version(tmp, separator=separator)
        if not tmp.exists():
            return tmp



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['organize', 'organize.actions', 'organize.filters']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.1,<6.0',
 'Send2Trash>=1.5,<2.0',
 'appdirs>=1.4,<2.0',
 'clint>=0.5.1,<0.6.0',
 'colorama>=0.4.1,<0.5.0',
 'docopt>=0.6.2,<0.7.0']

extras_require = \
{':python_version < "3.4"': ['pathlib2==2.3.0'],
 ':python_version < "3.5"': ['typing>=3.6,<4.0', 'pathlib2>=2.3.3,<3.0.0']}

entry_points = \
{'console_scripts': ['organize = organize.main:main']}

setup_kwargs = {
    'name': 'organize-tool',
    'version': '1.5.1',
    'description': 'The file management automation tool',
    'long_description': '.. image:: https://github.com/tfeldmann/organize/raw/master/docs/images/organize.svg?sanitize=true\n\n.. image:: https://readthedocs.org/projects/organize/badge/?version=latest\n  :target: https://organize.readthedocs.io/en/latest/?badge=latest\n  :alt: Documentation Status\n\n.. image:: https://travis-ci.org/tfeldmann/organize.svg?branch=master\n    :target: https://travis-ci.org/tfeldmann/organize\n\norganize\n========\n**The file management automation tool.**\n\nInstall via pip (requirement: Python 3.4+):\n\nOn macOS / Windows:\n``$ pip3 install organize-tool``\n\nOn Linux:\n``$ sudo pip3 install organize-tool``\n\n\nWhy you might find this useful\n------------------------------\nYour desktop is a mess? You cannot find anything in your downloads and\ndocuments? Sorting and renaming all these files by hand is too tedious?\nTime to automate it once and benefit from it forever.\n\n*organize* is a command line, open-source alternative to apps like Hazel (macOS)\nor File Juggler (Windows).\n\nIn your shell, run ``$ organize config`` to edit the configuration:\n\n- ``config.yaml``:\n\n  .. code-block:: yaml\n\n      rules:\n        # move screenshots into "Screenshots" folder\n        - folders:\n            - ~/Desktop\n          filters:\n            - filename:\n                startswith: \'Screen Shot\'\n          actions:\n            - move: ~/Desktop/Screenshots/\n\n        # move incomplete downloads older > 30 days into the trash\n        - folders:\n            - ~/Downloads\n          filters:\n            - extension:\n                - download\n                - crdownload\n                - part\n            - lastmodified:\n                days: 30\n                mode: older\n          actions:\n            - trash\n\n(alternatively you can run ``$ organize config --path`` to see the full path to\nyour ``config.yaml``)\n\n``$ organize run`` will now...\n\n- move all your screenshots from your desktop a "Screenshots" subfolder\n  (the folder will be created if it does not exist)\n- put all incomplete downloads older than 30 days into the trash\n\nIt is that easy.\n\nFeeling insecure? Run ``$ organize sim`` to see what would happen without\ntouching your files.\n\nBut there is more. You want to rename / copy files, run custom shell- or python\nscripts, match filenames with regular expressions or use placeholder variables?\n`organize` has you covered.\n\nHave a look at the full documentation at https://organize.readthedocs.io/.\n\n\nAdvanced usage example\n----------------------\nThis example shows some advanced features like placeholder variables, pluggable\nactions and recursion through subfolders:\n\n.. code-block:: yaml\n\n    rules:\n      - folders: \'~/Documents\'\n        subfolders: true\n        filters:\n          - extension:\n              - pdf\n              - docx\n          - lastmodified\n        actions:\n          - move: \'~/Documents/{extension.upper}/{lastmodified.year}-{lastmodified.month:02}/\'\n          - shell: \'open "{path}"\'\n\nGiven we have two files in our ``~/Documents`` folder (or any of its subfolders)\nnamed ``script.docx`` from january 2018 and ``demo.pdf`` from december 2016 this will\nhappen:\n\n- ``script.docx`` will be moved to ``~/Documents/DOCX/2018-01/script.docx``\n- ``demo.pdf`` will be moved to ``~/Documents/PDF/2016-12/demo.pdf``\n- The files will be opened (``open`` command in macOS) from their new location.\n\n\nFunctionality\n-------------\n\n**Select files by** (filters):\n\n- Extension\n- Regular expression\n- Last modified date (newer, older)\n- Filename (startswith, endswith, contains)\n\n**Organize your files** (actions):\n\n- Move files\n- Copy files\n- Rename files in place\n- Run shell command\n- Run inline Python code\n- Move into Trash\n- Print something to the console\n\nIf you miss a feature please file an issue. Pull requests welcome!\n\n\nCommand line interface\n----------------------\n::\n\n  The file management automation tool.\n\n  Usage:\n      organize sim [--config-file=<path>]\n      organize run [--config-file=<path>]\n      organize config [--open-folder | --path | --debug] [--config-file=<path>]\n      organize list\n      organize --help\n      organize --version\n\n  Arguments:\n      sim             Simulate a run. Does not touch your files.\n      run             Organizes your files according to your rules.\n      config          Open the configuration file in $EDITOR.\n      list            List available filters and actions.\n      --version       Show program version and exit.\n      -h, --help      Show this screen and exit.\n\n  Options:\n      -o, --open-folder  Open the folder containing the configuration files.\n      -p, --path         Show the path to the configuration file.\n      -d, --debug        Debug your configuration file.\n\n  Full documentation: https://organize.readthedocs.io\n',
    'author': 'Thomas Feldmann',
    'author_email': 'mail@tfeldmann.de',
    'url': 'https://github.com/tfeldmann/organize',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)



class Filter:
    def matches(self, path):
        """ Return True if filter applies to path. False otherwise """
        raise NotImplementedError

    def parse(self, path):
        """ Return an dict of parsed file properties (optional) """
        return {}

    def __str__(self):
        """ Return filter name and properties """
        return self.__class__.__name__

    def __repr__(self):
        return "<%s>" % str(self)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



import logging
import os
import shutil

from organize.utils import Path, fullpath, find_unused_filename

from .action import Action
from .trash import Trash


class Copy(Action):

    """
    Copy a file to a new location.
    If the specified path does not exist it will be created.

    :param str dest:
        The destination where the file should be copied to.
        If `dest` ends with a slash / backslash, the file will be copied into
        this folder and keep its original name.

    :param bool overwrite:
        specifies whether existing files should be overwritten.
        Otherwise it will start enumerating files (append a counter to the
        filename) to resolve naming conflicts. [Default: False]

    :param str counter_separator:
        specifies the separator between filename and the appended counter.
        Only relevant if **overwrite** is disabled. [Default: ``\' \'``]

    Examples:
        - Copy all pdfs into `~/Desktop/somefolder/` and keep filenames

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - extension: pdf
                actions:
                  - copy: '~/Desktop/somefolder/'

        - Use a placeholder to copy all .pdf files into a "PDF" folder and all .jpg
          files into a "JPG" folder. Existing files will be overwritten.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - extension:
                      - pdf
                      - jpg
                actions:
                  - copy:
                      dest: '~/Desktop/{extension.upper}/'
                      overwrite: true

        - Copy into the folder `Invoices`. Keep the filename but do not
          overwrite existing files. To prevent overwriting files, an index is
          added to the filename, so `somefile.jpg` becomes `somefile 2.jpg`.
          The counter separator is `' '` by default, but can be changed using
          the `counter_separator` property.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop/Invoices
                filters:
                  - extension:
                      - pdf
                actions:
                  - copy:
                      dest: '~/Documents/Invoices/'
                      overwrite: false
                      counter_separator: '_'
    """

    def __init__(self, dest: str, overwrite=False, counter_separator=" "):
        self.dest = dest
        self.overwrite = overwrite
        self.counter_separator = counter_separator
        self.log = logging.getLogger(__name__)

    def run(self, attrs: dict, simulate: bool):
        path = attrs["path"]
        basedir = attrs["basedir"]

        expanded_dest = self.fill_template_tags(self.dest, attrs)
        # if only a folder path is given we append the filename to have the full
        # path. We use os.path for that because pathlib removes trailing slashes
        if expanded_dest.endswith(("\\", "/")):
            expanded_dest = os.path.join(expanded_dest, path.name)

        new_path = fullpath(expanded_dest)
        if new_path.exists() and not new_path.samefile(path):
            if self.overwrite:
                self.print("File already exists")
                Trash().run({"path": new_path}, simulate=simulate)
            else:
                new_path = find_unused_filename(
                    path=new_path, separator=self.counter_separator
                )

        self.print('Copy to "%s"' % new_path)
        if not simulate:
            self.log.info("Creating folder if not exists: %s", new_path.parent)
            new_path.parent.mkdir(parents=True, exist_ok=True)
            self.log.info('Copying "%s" to "%s"', path, new_path)
            shutil.copy2(src=str(path), dst=str(new_path))

        # the next actions should handle the original file
        return None

    def __str__(self):
        return "Copy(dest=%s, overwrite=%s)" % (self.dest, self.overwrite)
